% This function returns the coefficient estimates using MPA and individual 
% estimation based on fixed window of training sample.
%
% Preliminary model screening is based on C-Lasso.
%
% Input: 
% x = explanatory variables
% y = dependent variable
% N,T = sample size
% K = number of regressors
% pred = indicating the variable of interest (coefficient estimation or prediction)
% boot = bootstrap indicator
% lambda_tmp = tuning parameters
% outsample_ratio = out of sample ratio for forecasting 
%
% Output:
% bmpa_bh = estimated slope coefficient of MPA (between heteroskedasticity)(pred = 0)
 % stdmpa = bootstrap standard deviation of MPA estimator (only valid when boot = 1 pred = 0)

function [bmpa_bh,stdmpa] = mpa_est(x,y,N,T,K,pred,boot,lambda_tmp,outsample_ratio)

outsampT            = round(outsample_ratio*T);
insampT             = T-outsampT;
IT                  = insampT;
ix                  = x(:,1:insampT,:);
iy                  = y(:,1:insampT);
ox                  = x(:,insampT+1:end,:);
oy                  = y(:,insampT+1:end);
Y                   = reshape(iy',N*IT,1);                                    
X                   = zeros(N*IT,N*K);
%% Individual estimation
% --- Estimate individual time series
for i = 1:N
    Z               = reshape(ix(i,:,:),IT,K);
    YY              = iy(i,:)';
    estb            = (Z'*Z)\(Z'*YY);
    resid(:,i)      = YY-Z*estb;
    hatbeta(:,i)    = estb;                                                 % estimates in individual regression
    u(:,i)          = hatbeta(:,i);
    hatsigma2(i)    = resid(:,i)'*resid(:,i)/(IT-K);                        % square error estimate in indiv regression
    varu(:,:,i)     = hatsigma2(i)*inv(Z'*Z);
end
unbeta              = reshape(hatbeta,N*K,1);                               % unrestricted least square estimator
beta_hat0           = hatbeta';

%--- Variance-covariance 
Vhomo               = zeros(N*K,N*K);                                       
Vbh                 = zeros(N*K,N*K);                                       
Vch                 = zeros(N*K,N*K);                                       

for i = 1:N
    xi                               = reshape(ix(i,:,:),IT,K);
    X((i-1)*IT+1:i*IT,(i-1)*K+1:i*K) = xi;
    Vbh((i-1)*K+1:i*K,(i-1)*K+1:i*K) = varu(:,:,i);                         % between-individual heteroskedastic error
    for t = 1:IT
        xit         = reshape(ix(i,t,:),K,1);
        u2xx(i,t)   = resid(t,i)^2*xit'*xit;
    end
    % completely (conditional) heteroskedastic error structure
    Vch((i-1)*K+1:i*K,(i-1)*K+1:i*K)    = 1/(IT*(IT-K))*inv(xi'*xi/IT)*sum(u2xx(i,:))*inv(xi'*xi/IT);    
    SX((i-1)*IT+1:i*IT,(i-1)*K+1:i*K)     = X((i-1)*IT+1:i*IT,(i-1)*K+1:i*K)/sqrt(hatsigma2(i)); % scaled regressors (for heteroskecastic BMA)
    SY((i-1)*IT+1:i*IT,1)                 = Y((i-1)*IT+1:i*IT)/sqrt(hatsigma2(i));               % scaled dependent variable (for heteroskecastic BMA)
end
Vhomo               = inv(X'*X)*((Y-X*unbeta)'*(Y-X*unbeta)/((N*IT)-(N*K))); % homoskedastic errors

%% Grouping using C-Lasso
a0                  = unique(beta_hat0,'rows');
M0                  = size(a0,1);
M                   = round(N/2);

sequence_lambda     = lambda_tmp /( IT^(1/3) );
lambda              = sequence_lambda * var(Y);
ssr                 = ones(1,M) * realmax;
g                   = ones(M,N);
for m = 1:M
    [~, g(m,:), ssr(m)]  = est_group_Classo(ix,iy,beta_hat0,N,IT,K,m,lambda);    
end
[~, bestG]                 = min( log(ssr) + 2/3*(N*IT)^(-.5)*K*(1:M));


%% Create restriction matrix with grouping from Classo
indexsort               = repmat(1:N,K,1);
gid                     = ones(M,N,K);
for k = 1:K
    for m = 1:M
        gid(m,:,k)      = g(m,:);
        RR(:,:,k,m)     = zeros(N-1,N*K);                                   % Restriction matrix with Redundant rows
        r               = 1;
        i               = 1;
        j               = 1;
        % recover the original order
        while i <= (N-1) & j <= N
            j = i+1;
            while j <= N
                if gid(m,i,k) == gid(m,j,k)
                    RR(r,(indexsort(k,i)-1)*K+k,k,m) = 1;
                    RR(r,(indexsort(k,j)-1)*K+k,k,m) = -1;
                    j                                = j+1;
                    r                                = r+1;
                else
                    for v=1:j-1
                        if gid(m,v,k)==gid(m,j,k)
                            RR(r,(indexsort(k,v)-1)*K+k,k,m) = 1;
                            RR(r,(indexsort(k,j)-1)*K+k,k,m) = -1;
                            r=r+1;
                            break;
                        end
                    end
                    i = j;
                    j = i+1;
                end
            end
        end
    end
end

%% Estimation under each restriction
minaic  = 1000000;
minbic  = 1000000;
if pred == 1
    A           = X'*X;
elseif pred == 0
    A           = eye(N*K,N*K);
end
for m = 1:M
    SR               = zeros((N-1)*K,N*K);
    for k = 1:K
        SR((k-1)*(N-1)+1:k*(N-1),:) = RR(:,:,k,m);
    end
    SR               = SR(max(SR~=0,[],2),:);                                % Simplied Restriction matrix under m1,...,m5 grouping
    P                = eye(N*K,N*K)-(X'*X)^(-1)*SR'/(SR/(X'*X)*SR')*SR;
    rebeta(:,m)      = P*unbeta;                                             % restricted least square estimator under grouping m
    Qhomo(m)         = 2*trace(P'*A*Vhomo);
    Qbh(m)           = 2*trace(P'*A*Vbh);
    Qch(m)           = 2*trace(P'*A*Vch);
end
SM                  = size(rebeta,2);
bestb_classo        = rebeta(:,bestG);                                      % best c-lasso estimate

%% Model averaging
% --- Mallow model averaging
D           = rebeta - repmat(unbeta,1,SM);
a1          = 2.*D'*A*D;
a1          = (a1+a1')/2;
a2homo      = Qhomo;
a2bh        = Qbh;
a2ch        = Qch;
w0          = ones(SM,1)/SM;
options     = optimset('Display','off','Algorithm','interior-point-convex');
% whomo       = quadprog(a1,a2homo,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
wbh         = quadprog(a1,a2bh,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
wch         = quadprog(a1,a2ch,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
% whomo       = whomo.*(whomo>0);
% whomo       = whomo/sum(whomo)';
wbh         = wbh.*(wbh>0);
wbh         = wbh/sum(wbh)';
wch         = wch.*(wch>0);
wch         = wch/sum(wch)';
% bmpa_homo   = rebeta*whomo;
bmpa_bh     = rebeta*wbh;
bmpa_ch     = rebeta*wch;

%% Bootstrap standard deviation
if boot == 1
B                       = 200;
for b = 1:B
    % --- Random sampling
    samp                = [1:1:N];
    indr                = randsample(samp,N,1);
    indr                = sort(indr);
    for i = 1:N   
        bx(i,:,:)       = x(indr(i),:,:);        
        by(i,:)         = y(indr(i),:);    
    end
    for i = 1:N
        Z               = reshape(bx(i,:,:),T,K);
        YY              = by(i,:)';
        estb            = (Z'*Z)\(Z'*YY);
        resid(:,i)      = YY-Z*estb;
        hatbeta(:,i)    = estb;                                                 % estimates in individual regression
        u(:,i)          = hatbeta(:,i);
        hatsigma2(i)    = resid(:,i)'*resid(:,i)/(T-K);                         % square error estimate in indiv regression
        varu(:,:,i)     = hatsigma2(i)*inv(Z'*Z);
    end
    unbeta              = reshape(hatbeta,N*K,1);                               % unrestricted least square estimator
    beta_hat0           = hatbeta';

    %--- Variance-covariance 
    Vhomo               = zeros(N*K,N*K);                                       
    Vbh                 = zeros(N*K,N*K);                                       
    Vch                 = zeros(N*K,N*K);                                       

    for i = 1:N
        xi                               = reshape(bx(i,:,:),T,K);
        X((i-1)*T+1:i*T,(i-1)*K+1:i*K) = xi;
        Vbh((i-1)*K+1:i*K,(i-1)*K+1:i*K) = varu(:,:,i);                         % between-individual heteroskedastic error
        for t = 1:T
            xit         = reshape(bx(i,t,:),K,1);
            u2xx(i,t)   = resid(t,i)^2*xit'*xit;
        end
        % completely (conditional) heteroskedastic error structure
        Vch((i-1)*K+1:i*K,(i-1)*K+1:i*K) = 1/(T*(T-K))*inv(xi'*xi/T)*sum(u2xx(i,:))*inv(xi'*xi/T);        
    end
    Vhomo               = inv(X'*X)*((Y-X*unbeta)'*(Y-X*unbeta)/((N*T)-(N*K))); % homoskedastic errors

%% Grouping using C-Lasso
    a0                  = unique(beta_hat0,'rows');
    M0                  = size(a0,1);
    M                   = round(N/2);

    sequence_lambda     = lambda_tmp /(IT^(1/3));
    lambda              = sequence_lambda * var(Y);
    ssr                 = ones(1,M) * realmax;
    g                   = ones(M,N);
    for m = 1:M
        [~, g(m,:), ssr(m)]  = est_group_Classo(bx,by,beta_hat0,N,T,K,m,lambda);    
    end
    [~, bestG]                 = min( log(ssr) + 2/3*(N*T)^(-.5)*K*(1:M));

%% Create restriction matrix with grouping from Classo
    indexsort               = repmat(1:N,K,1);
    gid                     = ones(M,N,K);
    for k = 1:K
        for m = 1:M
            gid(m,:,k)      = g(m,:);
            RR(:,:,k,m)     = zeros(N-1,N*K);                                   % Restriction matrix with Redundant rows
            r               = 1;
            i               = 1;
            j               = 1;
            % recover the original order
            while i <= (N-1) & j <= N
                j = i+1;
                while j <= N
                    if gid(m,i,k) == gid(m,j,k)
                        RR(r,(indexsort(k,i)-1)*K+k,k,m) = 1;
                        RR(r,(indexsort(k,j)-1)*K+k,k,m) = -1;
                        j                                = j+1;
                        r                                = r+1;
                    else
                        for v=1:j-1
                            if gid(m,v,k)==gid(m,j,k)
                                RR(r,(indexsort(k,v)-1)*K+k,k,m) = 1;
                                RR(r,(indexsort(k,j)-1)*K+k,k,m) = -1;
                                r=r+1;
                                break;
                            end
                        end
                        i = j;
                        j = i+1;
                    end
                end
            end
        end
    end

    %% Estimation under each restriction
    minaic  = 1000000;
    minbic  = 1000000;
    if pred == 1
        A           = X'*X;
    elseif pred == 0
        A           = eye(N*K,N*K);
    end
    for m = 1:M
        SR               = zeros((N-1)*K,N*K);
        for k = 1:K
            SR((k-1)*(N-1)+1:k*(N-1),:) = RR(:,:,k,m);
        end
        SR               = SR(max(SR~=0,[],2),:);                                % Simplied Restriction matrix under m1,...,m5 grouping
        P                = eye(N*K,N*K)-(X'*X)^(-1)*SR'/(SR/(X'*X)*SR')*SR;
        rebeta(:,m)      = P*unbeta;                                             % restricted least square estimator under grouping m
        Qhomo(m)         = 2*trace(P'*A*Vhomo);
        Qbh(m)           = 2*trace(P'*A*Vbh);
        Qch(m)           = 2*trace(P'*A*Vch);
    
        % --- Information criterion
        loglikelihood(m) = -(N*T)*log(2*pi)/2-(N*T)/2*log((Y-X*rebeta(:,m))'*(Y-X*rebeta(:,m))/(N*T))-(N*T)/2;
        aic(m)           = 2*(K*N-size(SR,1))-2*loglikelihood(m);
        bic(m)           = log(N*T)*(K*N-size(SR,1))-2*loglikelihood(m);
    
        % --- Minimum AIC
        if aic(m) < minaic
            minaic  = aic(m);
            maic    = m;
            betaaic = rebeta(:,m);
        end
    
        % --- Minimum BIC
        if bic(m) < minbic
            minbic  = bic(m);
            mbic    = m;
            betabic = rebeta(:,m);
        end    
    end
    SM                  = size(rebeta,2);
    bestb_classo        = rebeta(:,bestG);                                      % best c-lasso estimate

%% Model averaging
% --- Mallow model averaging
    D           = rebeta - repmat(unbeta,1,SM);
    a1          = 2.*D'*A*D;
    a1          = (a1+a1')/2;
    a2homo      = Qhomo;
    a2bh        = Qbh;
    a2ch        = Qch;
    w0          = ones(SM,1)/SM;
    options     = optimset('Display','off','Algorithm','interior-point-convex');
%     whomo       = quadprog(a1,a2homo,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
    wbh         = quadprog(a1,a2bh,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
%     wch         = quadprog(a1,a2ch,zeros(1,SM),0,ones(1,SM),1,zeros(SM,1),ones(SM,1),w0,options);
%     whomo       = whomo.*(whomo>0);
%     whomo       = whomo/sum(whomo)';
    wbh         = wbh.*(wbh>0);
    wbh         = wbh/sum(wbh)';
    wch         = wch.*(wch>0);
    wch         = wch/sum(wch)';
    bmpa_bh(:,b)     = rebeta*wbh;
%     bmpa_homo(:,b)   = rebeta*whomo;
%     bmpa_ch(:,b)     = rebeta*wch;
b
end
stdmpa              = std(bmpa_bh'); 
b_mpa               = mean(bmpa_bh');
confint_mpa         = [prctile(bmpa_bh',5); prctile(bmpa_bh',95)];
end


